 /*************************************************************************
  outer.c
   Mijan Huq and Reid Guenther 1998
   Routine to compute values for option price for the max of many assets.
   From H.Johnson, "Options on the Maximum or minimum of several assets",
   Journal of Financial and Quantitative Analysis, Vol 22, No. 3, Sept
   1997, p 277-283..
  *************************************************************************/
 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
 /* - - - - - - - - - -prototype for cum normal dist- - - - - - - - - - */
 void mecdf_(int *N_underliers,double *d,double *rho,
            double *probability,int *ier);
 /*
  *
  *
  *
  */
 /*
   swaprc(...)
   Does appropriate row-column swaps as inferred from Johnson...
   This routine is needed by MaxOfnAssets...
   A := incoming matrix
   x := returning components of matrix in packed form for mecdf
 */
 void swaprc(double **A, double *x, int it, int N){
   int i, j, k;

 #ifdef TRACE
   for(i=0;i<N;i++){
     for(j=0;j<N;j++){
       printf("%lf \t",A[i][j]);
     }
     printf("\n");
   }
 #endif /* TRACE */

   /*- - - - -  swap columns - - - - - */
   /* copy over current column */
   for(i=0; i< N;i++){
     x[i] = A[i][it];
   }
   /* move all columns < it up one */
   for(i=it-1;i>=0;i--){
     printf("moving column %d \n", i);
     for(j=0; j<N; j++){
       A[j][i+1] = A[j][i];
     }
   }
   /* store it column in the 0 spot */
   for(j=0; j<N; j++){
     A[j][0] = x[j];
   }

 #ifdef TRACE
   for(i=0;i<N;i++){
     for(j=0;j<N;j++){
       printf("%lf \t",A[i][j]);
     }
     printf("\n");
   }
 #endif /* TRACE */

   /*- - - - -  swap rows - - - - - */
   /* copy over current rows */
   for(i=0; i< N;i++){
     x[i] = A[it][i];
   }
   /* move all rows < it up one */
   for(i=it-1;i>=0;i--){
     for(j=0; j<N; j++){
       A[i+1][j] = A[i][j];
     }
   }
   /* store it row in the 0 spot */
   for(j=0; j<N; j++){
     A[0][j] = x[j];
   }

 #ifdef TRACE
   for(i=0;i<N;i++){
     for(j=0;j<N;j++){
       printf("%lf \t",A[i][j]);
     }
     printf("\n");
   }
 #endif /* TRACE */

   /* - - - - - Extract off-diagonals - - - - - */
   k=0;
   for(i=0;i<N;i++){
     for(j=i+1;j<N;j++){
       x[k] = A[i][j];
       k++;
     }
   }

 } /* end swaprc */

 /*-----------------------------------------------------------------------
   MaxOfnAssets(...)
   Computes the option price for the max of n assets.
   S := vector of stock prices.
   X := strike price.
   r := risk-free interest rate.
   delta:= dividend rate.
   cov := covariance matrix.
   tau := time
   N := Number of underliers.
 */
 double MaxOfnAssets(double *S, double X, double r, double *delta, double **cov, double tau, int N)
 {
   int i,j,k, ier;
   double cmax, Nn;
   double **sigma, ***rho, **d1p, *d1, *d2, **tmp;
   double *nd, *nrho; /* arrays to pass as arguments to mecdf */

   /* - - - - - - - - - - Allocate memory  - - - - - - - - - -*/
 #ifdef TRACE
   printf("Allocating memory\n");
 #endif /* TRACE */

   for(i=0;i<N;i++){
     printf("===> S[%d] = %lf\n",i,S[i]);
     S[i] *= exp(-delta[i]*tau);
   }

   d1 = (double *)malloc(N * sizeof(double));
   d2 = (double *)malloc(N * sizeof(double));

   nd = (double *)malloc(N * sizeof(double));
   nrho = (double *)malloc(N*(N-1)/2 * sizeof(double));

   rho = (double ***)malloc(N * sizeof(double **));
   tmp = (double **)malloc(N * sizeof(double *));
   sigma = (double **)malloc(N * sizeof(double *));
   d1p = (double **)malloc(N * sizeof(double *));

  

 /*
   if( *d1p == NULL || d1 == NULL || d2 == NULL || **rho == NULL ||
       *sigma == NULL || nd == NULL || nrho == NULL){
     perror("MaxOfnAssets: Cannot allocate memory");
     exit(1);
   }
 */

   for(i=0;i<N;i++){
     rho[i] = (double **)malloc(N * sizeof(double *));
     for(j=0;j<N;j++){
       rho[i][j] = (double *)malloc(N * sizeof(double));
     }
     tmp[i] = (double *)malloc(N * sizeof(double));
     sigma[i] = (double *)malloc(N * sizeof(double));
     d1p[i] = (double *)malloc(N * sizeof(double));
   }
   /* - - - - - - - - - Compute sigma  - - - - - - - - - -- */
 #ifdef TRACE
   printf("Computing  sigma\n");
 #endif /* TRACE */
   /* First set up sigma_ij on top of page 280 */
   for(i=0;i<N;i++){
     for(j=0;j<N;j++){
        /* For i not equal to j */
        sigma[i][j] = cov[i][i] - 2.0 * cov[i][j] * sqrt(cov[i][i]) *
                      sqrt(cov[j][j]) + cov[j][j];
     }
        sigma[i][i] = (cov[i][i]);
   }
   for(i=0;i<N;i++){
     for(j=0;j<N;j++){
       sigma[i][j] = sqrt(sigma[i][j]);
     }
   }

   /* - - - - - - - - - - compute d1 primed - - - - - - - - - - */
 #ifdef TRACE
   printf("Computing  d1p\n");
 #endif /* TRACE */
   for(i=0; i<N; i++){
     for(j=0; j<N; j++){
       d1p[i][j] = log(S[i] / S[j]) +
                   0.5 * sigma[i][j]* sigma[i][j] * tau;
       d1p[i][j] /= sigma[i][j] * sqrt(tau);
     }
   }

   /* - - - - - - - - - - compute d2 - - - - - - - - - - */
 #ifdef TRACE
   printf("Computing  d2\n");
 #endif /* TRACE */
   for(i=0; i<N; i++){
     d2[i] = log(S[i]/X) + ( r - 0.5 * cov[i][i]) * tau;
     d2[i] /= sqrt(cov[i][i]) * sqrt(tau);
   }
   /* - - - - - - - - - - compute d1 - - - - - - - - - - */
 #ifdef TRACE
   printf("Computing  d1\n");
 #endif /* TRACE */
   for(i=0; i<N; i++){
     d1[i] = d2[i] + sqrt(cov[i][i]) * sqrt(tau);
   }
   /* - - - - - - - - - - compute rho - - - - - - - - - - */
 #ifdef TRACE
   printf("Computing  rho\n");
 #endif /* TRACE */
   for(i=0; i<N; i++){
     for(j=0; j<N; j++){
       for(k=0; k<N; k++){
         if(i != j){
           rho[i][j][k] = cov[i][i] - cov[i][j] * sqrt(cov[i][i]) *
                         sqrt(cov[j][j]) -  cov[i][k] * sqrt(cov[i][i]) *
                         sqrt(cov[k][k]) +  cov[j][k] * sqrt(cov[j][j]) *
                         sqrt(cov[k][k]);
           rho[i][j][k] /= sigma[i][j] * sigma[i][k];
         } else {
           rho[i][i][k] = (sqrt(cov[i][i]) - cov[i][k] *
                           sqrt(cov[k][k]) ) / sigma[i][k];
         }
       }
     }
   }

   /* - - - - - Now form each part of the positive parts of cmax - - - - - */
 #ifdef TRACE
   printf("Form  arguments to cumulative normal distribution\n");
 #endif /* TRACE */
   cmax = 0.0;
   for(i=0; i<N; i++){ /* index for sum over underliers */
     /* load in vector D */
     nd[0] = d1[i];
     for(j=1;j<N;j++){
       nd[j] = d1p[i][j];
     }

     /* - - - - - carry out array op's on rho and form rho - - - - - */
     /* Copy over matrix for underlier i */
 #ifdef TRACE
   printf("Copying over rho matrix to be swapped around\n");
 #endif /* TRACE */
     for(j=0;j<N;j++){
       for(k=0;k<N;k++){
         tmp[j][k] = rho[i][j][k];
       }
     }

     /* perform matrix row-column swap */
     swaprc(tmp, nrho,i,N);

     /*  - - - - -compute probability  - - - - -*/
    mecdf_(&N, nd, nrho, &Nn, &ier);

     /*  - - - - -add to cmax  - - - - -*/
     cmax += S[i] * Nn;

   } /* end loop i */

   /*  - - - - -Now form the negative part of cmax  - - - - -*/
   /* first form packed version of covariance matrix */
   k=0;
   for(i=0;i<N;i++){
     for(j=i+1;j<N;j++){
       nrho[k] = cov[i][j];
     }
     nd[i] = -d2[i];
   }
   /* compute probabilty */
   mecdf_(&N, nd, nrho, &Nn, &ier);

   /* Subtract from cmax */
   cmax -= X*exp(-r * tau) * (1.0 - Nn);

   for(i=0;i<N;i++){
     S[i] *= exp(delta[i]*tau);
   }
   /* Free memory */
   free(d1p);
   free(d1);
   free(d2);
   free(nrho);
   free(nd);
   for(i=0;i<N;i++){
     free(rho[i]);
     free(tmp[i]);
     free(sigma[i]);
   }
   free(rho);
   free(sigma);
   free(tmp);
   return(cmax);
 } /* end MaxOfnAssets */
